import unittest
from pygame_gridcalculator.gridcalculator import GridCalculator, \
    GridCalculatorException


class TestGridCalculator(unittest.TestCase):
    def setUp(self) -> None:
        self.test_grid = GridCalculator(100, 100, 5, 5)
        self.test_grid2 = GridCalculator(10, 20, 5, 10)

    def test_init_grid(self) -> None:
        """Test initialising grid works as expected"""
        self.assertEqual(self.test_grid.size, (5, 5))
        self.assertEqual(self.test_grid.pixel_size, (100, 100))
        self.assertEqual(self.test_grid2.size, (5, 10))
        self.assertEqual(self.test_grid2.pixel_size, (10, 20))

    def test_init_grid_with_pixel_start_positions(self) -> None:
        """Test initialising grid with start pixel positions works"""
        new_grid = GridCalculator(300, 300, 2, 2, 100, 100)
        self.assertEqual(new_grid.size, (2, 2))
        self.assertEqual(new_grid.pixel_size, (200, 200))

    def test_init_grid_error_pixel_end_left_too_low(self) -> None:
        """Test initialising grid with 0 sized pixel end left errors"""
        with self.assertRaises(GridCalculatorException) as err:
            GridCalculator(0, 100, 0, 0)
        self.assertEqual(str(err.exception),
                         "Pixel end left must be greater than 1")

    def test_init_grid_error_pixel_end_top_too_low(self) -> None:
        """Test initialising grid with 0 sized pixel end top errors"""
        with self.assertRaises(GridCalculatorException) as err:
            GridCalculator(100, 0, 0, 0)
        self.assertEqual(str(err.exception),
                         "Pixel end top must be greater than 1")

    def test_init_grid_error_grid_width_too_low(self) -> None:
        """Test initialising grid with 0 sized grid width errors"""
        with self.assertRaises(GridCalculatorException) as err:
            GridCalculator(100, 100, 0, 10)
        self.assertEqual(str(err.exception),
                         "Grid width must be greater than 1")

    def test_init_grid_error_grid_height_too_low(self) -> None:
        """Test initialising grid with 0 sized grid height errors"""
        with self.assertRaises(GridCalculatorException) as err:
            GridCalculator(100, 100, 10, 0)
        self.assertEqual(str(err.exception),
                         "Grid height must be greater than 1")

    def test_init_grid_error_grid_width_bigger_than_end_left(self) -> None:
        """Test initialising grid with a pixel end left smaller than the grid
        width errors"""
        with self.assertRaises(GridCalculatorException) as err:
            GridCalculator(100, 100, 200, 10)
        self.assertEqual(str(err.exception),
                         "The grid width (200) cannot be greater than the "
                         "pixel end left (100)")

    def test_init_grid_error_grid_height_bigger_than_end_top(self) -> None:
        """Test initialising grid with a pixel end top smaller than the grid
        height errors"""
        with self.assertRaises(GridCalculatorException) as err:
            GridCalculator(100, 100, 10, 200)
        self.assertEqual(str(err.exception),
                         "The grid height (200) cannot be greater than the "
                         "pixel end top (100)")

    def test_init_grid_error_grid_start_left_bigger_than_end(self) -> None:
        """Test initialising grid with a pixel start left bigger than pixel end
         left errors"""
        with self.assertRaises(GridCalculatorException) as err:
            GridCalculator(100, 100, 10, 10, 150, 50)
        self.assertEqual(str(err.exception),
                         "The grid width (10) cannot be greater than the pixel"
                         " end left (-50)")

    def test_init_grid_error_grid_start_top_bigger_than_end(self) -> None:
        """Test initialising grid with a pixel start top bigger than pixel end
         top errors"""
        with self.assertRaises(GridCalculatorException) as err:
            GridCalculator(100, 100, 10, 10, 50, 101)
        self.assertEqual(str(err.exception),
                         "The grid height (10) cannot be greater than the "
                         "pixel end top (-1)")

    def test_update_grid(self) -> None:
        """Test updating the grid works as expected"""
        self.test_grid.update_grid(10, 10)
        self.assertEqual(self.test_grid.size, (10, 10))

    def test_update_grid_error_grid_width_too_low(self) -> None:
        """Test changing the grid to a width too low doesn't work"""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.update_grid(0, 2)
        self.assertEqual(str(err.exception),
                         "Grid width must be greater than 1")

    def test_update_grid_error_grid_height_too_low(self) -> None:
        """Test changing the grid to a height too low doesn't work"""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.update_grid(2, 0)
        self.assertEqual(str(err.exception),
                         "Grid height must be greater than 1")

    def test_update_pixel_positions(self) -> None:
        """Test updating the pixel positions works"""
        self.test_grid.update_pixel_positions(120, 120)
        self.assertEqual(self.test_grid.points_from_right(0), 120)
        self.assertEqual(self.test_grid.points_from_bottom(0), 120)

    def test_update_pixel_positions_error_end_left_smaller_than_grid(self):
        """Test changing the pixel end left to a width smaller than the grid
        doesn't work"""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.update_pixel_positions(4, 20)
        self.assertEqual(str(err.exception),
                         "The grid width (5) cannot be greater than the "
                         "pixel end left (4)")

    def test_update_pixel_positions_error_end_top_smaller_than_grid(self):
        """Test changing the pixel end top to a height smaller than the grid
        doesn't work"""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.update_pixel_positions(20, 3)
        self.assertEqual(str(err.exception),
                         "The grid height (5) cannot be greater than the "
                         "pixel end top (3)")

    def test_left_error_check_too_low(self) -> None:
        """Test the left error check."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid._error_check_left(-1)
        self.assertEqual(str(err.exception),
                         "The left point provided (-1) isn't in the "
                         "grid (0 - 5)")

    def test_left_error_check_too_high(self) -> None:
        """Test the left error check."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid._error_check_left(6)
        self.assertEqual(str(err.exception),
                         "The left point provided (6) isn't in the "
                         "grid (0 - 5)")

    def test_top_error_check_too_low(self) -> None:
        """Test the top error check."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid._error_check_top(-1)
        self.assertEqual(str(err.exception),
                         "The top point provided (-1) isn't in the "
                         "grid (0 - 5)")

    def test_top_error_check_too_high(self) -> None:
        """Test the left error check."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid._error_check_top(6)
        self.assertEqual(str(err.exception),
                         "The top point provided (6) isn't in the "
                         "grid (0 - 5)")

    def test_get_size(self) -> None:
        """Test the size of the grid is returned correctly."""
        self.assertEqual(self.test_grid.size, (5, 5))
        self.assertEqual(self.test_grid2.size, (5, 10))

    def test_top_point(self) -> None:
        """Test getting the top point returns the right value."""
        self.assertEqual(self.test_grid.top_point(0), 0)
        self.assertEqual(self.test_grid.top_point(1), 20)
        self.assertEqual(self.test_grid.top_point(2), 40)
        self.assertEqual(self.test_grid.top_point(3), 60)
        self.assertEqual(self.test_grid.top_point(4), 80)
        self.assertEqual(self.test_grid.top_point(5), 100)

        self.assertEqual(self.test_grid2.top_point(0), 0)
        self.assertEqual(self.test_grid2.top_point(1), 2)
        self.assertEqual(self.test_grid2.top_point(2), 4)
        self.assertEqual(self.test_grid2.top_point(3), 6)
        self.assertEqual(self.test_grid2.top_point(4), 8)
        self.assertEqual(self.test_grid2.top_point(5), 10)
        self.assertEqual(self.test_grid2.top_point(6), 12)
        self.assertEqual(self.test_grid2.top_point(7), 14)
        self.assertEqual(self.test_grid2.top_point(8), 16)
        self.assertEqual(self.test_grid2.top_point(9), 18)
        self.assertEqual(self.test_grid2.top_point(10), 20)

    def test_top_point_error(self) -> None:
        """Test getting the top point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.top_point(60)
        self.assertEqual(str(err.exception),
                         "The top point provided (60) isn't in the "
                         "grid (0 - 5)")

    def test_left_point(self) -> None:
        """Test getting the left point returns the right value."""
        self.assertEqual(self.test_grid.left_point(0), 0)
        self.assertEqual(self.test_grid.left_point(1), 20)
        self.assertEqual(self.test_grid.left_point(2), 40)
        self.assertEqual(self.test_grid.left_point(3), 60)
        self.assertEqual(self.test_grid.left_point(4), 80)
        self.assertEqual(self.test_grid.left_point(5), 100)

        self.assertEqual(self.test_grid2.left_point(0), 0)
        self.assertEqual(self.test_grid2.left_point(1), 2)
        self.assertEqual(self.test_grid2.left_point(2), 4)
        self.assertEqual(self.test_grid2.left_point(3), 6)
        self.assertEqual(self.test_grid2.left_point(4), 8)
        self.assertEqual(self.test_grid2.left_point(5), 10)

    def test_left_point_error(self) -> None:
        """Test getting the left point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.left_point(33)
        self.assertEqual(str(err.exception),
                         "The left point provided (33) isn't in the "
                         "grid (0 - 5)")

    def test_position(self) -> None:
        """Test getting the position returns the right value."""
        self.assertEqual(self.test_grid.position(1, 1), (20, 20))
        self.assertEqual(self.test_grid.position(3, 2), (60, 40))
        self.assertEqual(self.test_grid.position(4, 5), (80, 100))

        self.assertEqual(self.test_grid2.position(0, 0), (0, 0))
        self.assertEqual(self.test_grid2.position(2, 1), (4, 2))
        self.assertEqual(self.test_grid2.position(4, 5), (8, 10))
        self.assertEqual(self.test_grid2.position(5, 10), (10, 20))

    def test_position_error_left_point(self) -> None:
        """Test getting the left point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.position(9, 2)
        self.assertEqual(str(err.exception),
                         "The left point provided (9) isn't in the "
                         "grid (0 - 5)")

    def test_position_error_top_point(self) -> None:
        """Test getting the top point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.position(2, 8)
        self.assertEqual(str(err.exception),
                         "The top point provided (8) isn't in the "
                         "grid (0 - 5)")

    def test_height_gap(self) -> None:
        """Test getting the height gap returns the right value."""
        self.assertEqual(self.test_grid.height_gap(1, 2), 20)
        self.assertEqual(self.test_grid.height_gap(2, 4), 40)
        self.assertEqual(self.test_grid.height_gap(0, 5), 100)

    def test_height_gap_error_top_point1(self) -> None:
        """Test getting the first top point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.height_gap(-2, 3)
        self.assertEqual(str(err.exception),
                         "The top point provided (-2) isn't in the "
                         "grid (0 - 5)")

    def test_height_gap_error_top_point2(self) -> None:
        """Test getting the second top point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.height_gap(3, 10)
        self.assertEqual(str(err.exception),
                         "The top point provided (10) isn't in the "
                         "grid (0 - 5)")

    def test_height_gap_error_point1_higher_than_point2(self) -> None:
        """Test getting the second top point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.height_gap(4, 3)
        self.assertEqual(str(err.exception),
                         "top_point1 is greater than top_point2")

    def test_width_gap(self) -> None:
        """Test getting the width gap returns the right value."""
        self.assertEqual(self.test_grid.width_gap(3, 4), 20)
        self.assertEqual(self.test_grid.width_gap(0, 3), 60)
        self.assertEqual(self.test_grid.width_gap(0, 5), 100)

    def test_width_gap_error_left_point1(self) -> None:
        """Test getting the first left point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.height_gap(-2, 4)
        self.assertEqual(str(err.exception),
                         "The top point provided (-2) isn't in the "
                         "grid (0 - 5)")

    def test_width_gap_error_left_point2(self) -> None:
        """Test getting the second left point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.height_gap(0, 6)
        self.assertEqual(str(err.exception),
                         "The top point provided (6) isn't in the "
                         "grid (0 - 5)")

    def test_square(self) -> None:
        """Test getting the square returns the right values."""
        self.assertEqual(self.test_grid.square(0, 0, 2, 2), (40, 40))
        self.assertEqual(self.test_grid.square(1, 2, 2, 5), (20, 60))

    def test_square_error_left_start_invalid(self) -> None:
        """Test getting the left start point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.square(-4, 1, 2, 2)
        self.assertEqual(str(err.exception),
                         "The left point provided (-4) isn't in the "
                         "grid (0 - 5)")

    def test_square_error_top_start_invalid(self) -> None:
        """Test getting the top start point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.square(1, -2, 7, 2)
        self.assertEqual(str(err.exception),
                         "The top point provided (-2) isn't in the "
                         "grid (0 - 5)")

    def test_square_error_left_end_invalid(self) -> None:
        """Test getting the left end point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.square(1, 1, 7, 2)
        self.assertEqual(str(err.exception),
                         "The left point provided (7) isn't in the "
                         "grid (0 - 5)")

    def test_square_error_top_end_invalid(self) -> None:
        """Test getting the top end point returns an error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.square(1, 1, 3, 8)
        self.assertEqual(str(err.exception),
                         "The top point provided (8) isn't in the "
                         "grid (0 - 5)")

    def test_square_error_left_end_greater_than_left_start(self) -> None:
        """Test providing a left end larger than left start point returns an
        error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.square(4, 1, 3, 4)
        self.assertEqual(str(err.exception),
                         "left_start is greater than left_end")

    def test_square_error_top_end_greater_than_top_start(self) -> None:
        """Test providing a top end larger than top start point returns an
        error when invalid."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.square(1, 3, 3, 2)
        self.assertEqual(str(err.exception),
                         "top_start is greater than top_end")

    def test_points_from_left(self) -> None:
        """Test getting the points from left returns the right value."""
        self.assertEqual(self.test_grid.points_from_left(2), 40)
        self.assertEqual(self.test_grid.points_from_left(4), 80)

    def test_points_from_left_error(self) -> None:
        """Test providing an invalid points from left value doesn't work."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.points_from_left(7)
        self.assertEqual(str(err.exception),
                         "The left point provided (7) isn't in the "
                         "grid (0 - 5)")

    def test_points_from_top(self) -> None:
        """Test getting the points from top returns the right value."""
        self.assertEqual(self.test_grid.points_from_top(3), 60)
        self.assertEqual(self.test_grid.points_from_top(5), 100)

    def test_points_from_top_error(self) -> None:
        """Test providing an invalid points from top value doesn't work."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.points_from_top(8)
        self.assertEqual(str(err.exception),
                         "The top point provided (8) isn't in the "
                         "grid (0 - 5)")

    def test_points_from_right(self) -> None:
        """Test getting the points from right returns the right value."""
        self.assertEqual(self.test_grid.points_from_right(1), 80)
        self.assertEqual(self.test_grid.points_from_right(4), 20)

    def test_points_from_right_error(self) -> None:
        """Test providing an invalid points from right value doesn't work."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.points_from_right(6)
        self.assertEqual(str(err.exception),
                         "The left point provided (-1) isn't in the "
                         "grid (0 - 5)")

    def test_points_from_bottom(self) -> None:
        """Test getting the points from right returns the right value."""
        self.assertEqual(self.test_grid.points_from_bottom(2), 60)
        self.assertEqual(self.test_grid.points_from_bottom(4), 20)

    def test_points_from_bottom_error(self) -> None:
        """Test providing an invalid points from bottom value doesn't work."""
        with self.assertRaises(GridCalculatorException) as err:
            self.test_grid.points_from_bottom(7)
        self.assertEqual(str(err.exception),
                         "The top point provided (-2) isn't in the "
                         "grid (0 - 5)")


if __name__ == '__main__':
    unittest.main()



from pygame import Surface, draw


class GridCalculatorException(Exception):
    """An exception used for GridCalculator-based exceptions."""
    pass


class GridCalculator:
    """Create a GridCalculator to map the available pixels on the screen
       against a grid of the users specified size.

    Parameters:
        pixel_end_left (int): The right pixel point for use by the grid.
        pixel_end_top (int): The bottom pixel point for use by the grid.
        grid_width_max (int): The max width point of the grid (right border).
        grid_height_max (int): The max height point of the grid (bottom
                               border).
        pixel_start_left (int): (Optional) The left pixel point for use by the
                                grid (default = 0).
        pixel_start_top (int): (Optional) The top pixel point for use by the
                               grid (default = 0).
    """

    def __init__(self, pixel_end_left: int, pixel_end_top: int,
                 grid_width_max: int, grid_height_max: int,
                 pixel_start_left: int = 0, pixel_start_top: int = 0):
        # Create and calculate grid
        self._check_value_links_are_valid(pixel_end_left, pixel_end_top,
                                          grid_width_max, grid_height_max,
                                          pixel_start_left, pixel_start_top)
        self._pixel_end_left = pixel_end_left
        self._pixel_end_top = pixel_end_top
        self._grid_width_max = grid_width_max
        self._grid_height_max = grid_height_max
        self._pixel_start_left = pixel_start_left
        self._pixel_start_top = pixel_start_top

    @staticmethod
    def _check_value_links_are_valid(pixel_end_left: int, pixel_end_top: int,
                                     grid_width_max: int, grid_height_max: int,
                                     pixel_start_left: int,
                                     pixel_start_top: int) -> None:
        """ Checks values which are inherently linked are valid before trying
        to set them.

        Parameters:
            pixel_end_left (int): The far right point to base the grid on
                                 in pixels.
            pixel_end_top (int): The bottom to base the grid on in pixels.
            grid_width_max (int): The max width point of the grid (right
                                  border).
            grid_height_max (int): The max height point of the grid (bottom
                                   border).
            pixel_start_left (int): The left pixel point for use by the grid.
            pixel_start_top (int): The top pixel point for use by the grid.
        """
        if (pixel_end_left - pixel_start_left) < grid_width_max:
            raise GridCalculatorException("The grid width ({}) cannot be "
                                          "greater than the pixel end left "
                                          "({})".format(grid_width_max,
                                                        (pixel_end_left -
                                                         pixel_start_left)))
        if (pixel_end_top - pixel_start_top) < grid_height_max:
            raise GridCalculatorException("The grid height ({}) cannot be "
                                          "greater than the pixel end top "
                                          "({})".format(grid_height_max,
                                                        (pixel_end_top -
                                                         pixel_start_top)))

    def __repr__(self):
        return "GridCalculator(pixel range: left={}-{}, top={}-{};  " \
               "grid: width={}, height={})".format(
                    self._pixel_start_left, self._pixel_end_left,
                    self._pixel_start_top, self._pixel_end_top,
                    self._grid_width_max, self._grid_height_max)

    def _get_width_pixels(self, point_needed: float) -> int:
        """Calculate the width point based on the value provided."""
        width = (self._pixel_end_left -
                 self._pixel_start_left) / self._grid_width_max
        return self._pixel_start_left + int(width * point_needed)

    def _get_height_pixels(self, point_needed: float) -> int:
        """Calculate the height point based on the value provided."""
        width = (self._pixel_end_top -
                 self._pixel_start_top) / self._grid_height_max
        return self._pixel_start_top + int(width * point_needed)

    @property
    def _pixel_end_left(self) -> int:
        """Get pixel end left point for grid."""
        return self._pix_end_left

    @_pixel_end_left.setter
    def _pixel_end_left(self, pixel_end_left: int) -> None:
        """Set pixel end left point for grid."""
        if pixel_end_left < 1:
            raise GridCalculatorException("Pixel end left must be greater "
                                          "than 1")
        self._pix_end_left = pixel_end_left

    @property
    def _pixel_end_top(self) -> int:
        """Get pixel end top point for grid."""
        return self._pix_end_top

    @_pixel_end_top.setter
    def _pixel_end_top(self, pixel_end_top: int) -> None:
        """Set pixel end top point for grid."""
        if pixel_end_top < 1:
            raise GridCalculatorException("Pixel end top must be greater "
                                          "than 1")
        self._pix_end_top = pixel_end_top

    @property
    def _grid_width_max(self) -> int:
        """Get grid width."""
        return self._g_width

    @_grid_width_max.setter
    def _grid_width_max(self, grid_width_max: int) -> None:
        """Set grid width."""
        if grid_width_max < 1:
            raise GridCalculatorException("Grid width must be greater than 1")
        self._g_width = grid_width_max

    @property
    def _grid_height_max(self) -> int:
        """Get grid height."""
        return self._g_height

    @_grid_height_max.setter
    def _grid_height_max(self, grid_height_max: int) -> None:
        """Set grid height."""
        if grid_height_max < 1:
            raise GridCalculatorException("Grid height must be greater than 1")
        self._g_height = grid_height_max

    @property
    def _pixel_start_left(self) -> int:
        """Get pixel start left position."""
        return self._pix_start_left

    @_pixel_start_left.setter
    def _pixel_start_left(self, pixel_start_left: int) -> None:
        """Set pixel start left position."""
        if pixel_start_left > self._pixel_end_left:
            raise GridCalculatorException("The pixel start left position ({}) "
                                          "cannot be greater than pixel end "
                                          "left position ({})".format(
                                            pixel_start_left,
                                            self._pixel_end_left))
        elif pixel_start_left < 0:
            raise GridCalculatorException("The pixel start left position ({}) "
                                          "cannot be less than 0".format(
                                            pixel_start_left))
        self._pix_start_left = pixel_start_left

    @property
    def _pixel_start_top(self) -> int:
        """Get pixel start top position."""
        return self._pix_start_top

    @_pixel_start_top.setter
    def _pixel_start_top(self, pixel_start_top: int) -> None:
        """Set pixel start top position."""
        if pixel_start_top > self._pixel_end_top:
            raise GridCalculatorException("The pixel start top position ({}) "
                                          "cannot be greater than the pixel "
                                          "end top position ({})".format(
                                            pixel_start_top,
                                            self._pixel_end_top))
        elif pixel_start_top < 0:
            raise GridCalculatorException("The pixel start top position ({}) "
                                          "cannot be less than 0".format(
                                            pixel_start_top))
        self._pix_start_top = pixel_start_top

    @property
    def size(self) -> tuple:
        """Returns the size of the grid.

        Returns:
            tuple: The actual size of the grid (width, height)"""
        return self._grid_width_max, self._grid_height_max

    @property
    def pixel_size(self) -> tuple:
        """Returns the pixel size of the grid.

        Returns:
            tuple: The size of the grid in pixels (width, height)"""
        return (self._pixel_end_left - self._pixel_start_left), \
               (self._pixel_end_top - self._pixel_start_top)

    def _error_check_left(self, left_point: float) -> None:
        """Left point error checking, checks left_point is in grid."""
        if 0 > left_point or left_point > self._grid_width_max:
            raise GridCalculatorException(
                "The left point provided ({}) isn't in the grid "
                "(0 - {})".format(left_point, self._grid_width_max))

    def _error_check_top(self, top_point: float) -> None:
        """Top point error checking, checks top_point is in grid"""
        if 0 > top_point or top_point > self._grid_height_max:
            raise GridCalculatorException(
                "The top point provided ({}) isn't in the grid "
                "(0 - {})".format(top_point, self._grid_height_max))

    def update_grid(self, grid_width_max: int, grid_height_max: int) -> None:
        """Recalculates the grid based on the provided width and height.

        Parameters:
            grid_width_max (int): The max width point of the grid (right
                                  border).
            grid_height_max (int): The max height point of the grid (bottom
                                   border)."""
        self._check_value_links_are_valid(self._pixel_end_left,
                                          self._pixel_end_top,
                                          grid_width_max, grid_height_max,
                                          self._pixel_start_left,
                                          self._pixel_start_top)
        self._grid_width_max = grid_width_max
        self._grid_height_max = grid_height_max

    def update_pixel_positions(self, pixel_end_left: int, pixel_end_top: int,
                               pixel_start_left: int = 0,
                               pixel_start_top: int = 0) -> None:
        """Recalculates the pixel positions the grid is based on.

        Parameters:
            pixel_end_left (int): The right pixel point for use by the grid.
            pixel_end_top (int): The bottom pixel point for use by the grid.
            pixel_start_left (int): (Optional) The left pixel point for use by
                                    the grid (default = 0).
            pixel_start_top (int): (Optional) The top pixel point for use by
                                   the grid (default = 0)."""
        self._check_value_links_are_valid(pixel_end_left, pixel_end_top,
                                          self._grid_width_max,
                                          self._grid_height_max,
                                          pixel_start_left, pixel_start_top)
        self._pixel_end_left = pixel_end_left
        self._pixel_end_top = pixel_end_top
        self._pixel_start_left = pixel_start_left
        self._pixel_start_top = pixel_start_top

    def top_point(self, point: float) -> int:
        """Returns the pixel position of the top point specified.

        Parameters:
            point (float): The grid point required from the top of the grid.

        Returns:
            int: The pixel value represented by the top grid point."""
        self._error_check_top(point)
        return self._get_height_pixels(point)

    def left_point(self, point: float) -> int:
        """Returns the pixel position of the left point specified.

        Parameters:
            point (float): The grid point required from the left of the grid.

        Returns:
            int: The pixel value represented by the left grid point."""
        self._error_check_left(point)
        return self._get_width_pixels(point)

    def position(self, left_point: float, top_point: float) -> tuple:
        """Returns the pixel positions in the grid of the specified points in
        the format.

        Parameters:
            left_point(float): The grid point required from the left of the
                               grid
            top_point (float): The grid point required from the top of the grid

        Returns:
            tuple: The pixel values represented by the grid points (left, top)
        """
        self._error_check_left(left_point)
        self._error_check_top(top_point)
        return self.left_point(left_point), self.top_point(top_point)

    def height_gap(self, top_point1: float, top_point2: float) -> int:
        """Returns the pixel gap between two specified grid top points.

        Parameters:
            top_point1 (float): The first top grid point to measure from.
            top_point2 (float): The second top grid point to measure to.

        Returns:
            int: The number of pixels between the two top points."""
        self._error_check_top(top_point1)
        self._error_check_top(top_point2)
        if top_point1 > top_point2:
            raise GridCalculatorException("top_point1 is greater than "
                                          "top_point2")

        return self.top_point(top_point2) - self.top_point(top_point1)

    def width_gap(self, left_point1: float, left_point2: float) -> int:
        """Returns the pixel gap between two specified left points.

        Parameters:
            left_point1 (float): The first left grid point to measure from.
            left_point2 (float): The second left grid point to measure to.

        Returns:
            int: The number of pixels between the two left points."""
        self._error_check_left(left_point1)
        self._error_check_left(left_point2)
        if left_point1 > left_point2:
            raise GridCalculatorException("left_point1 is greater than "
                                          "left_point2")

        return self.left_point(left_point2) - self.left_point(left_point1)

    def square(self, left_start: float, top_start: float, left_end: float,
               top_end: float) -> tuple:
        """Returns the total size of a square in pixels based on the top left
        point provided and bottom right point provided.

        Parameters:
            left_start (float): The left grid point for the top left of the
                              square
            top_start (float): The top grid point for the top left of the
                               square
            left_end (float): The left grid point for the bottom right of the
                            square
            top_end (float): The top grid point for the bottom right of the
                           square

        Returns:
            tuple: The pixel width and height of the square outlined
                   (width, height)"""
        self._error_check_left(left_start)
        self._error_check_top(top_start)
        self._error_check_left(left_end)
        self._error_check_top(top_end)
        if left_start > left_end:
            raise GridCalculatorException("left_start is greater than "
                                          "left_end")
        if top_start > top_end:
            raise GridCalculatorException("top_start is greater than top_end")

        return self.width_gap(left_start, left_end), self.height_gap(top_start,
                                                                     top_end)

    def points_from_left(self, points: float) -> int:
        """Returns the pixel position from the specified number of points from
        the left border.

        Parameters:
            points (float): The number of grid points from the left border

        Returns:
            int: The left pixel value represented by the grid point."""
        self._error_check_left(points)
        return self.left_point(points)

    def points_from_top(self, points: float) -> int:
        """Returns the pixel position from the specified number of points from
        the top border.

        Parameters:
            points (float): The number of grid points from the left border

        Returns:
            int: The top pixel value represented by the grid point."""
        self._error_check_top(points)
        return self.top_point(points)

    def points_from_right(self, points: float) -> int:
        """Returns the pixel position from the specified number of points from
        the right border.

        Parameters:
            points (float): The number of grid points from the right border

        Returns:
            int: The left pixel value represented by the grid point."""
        self._error_check_left(self._grid_width_max - points)
        return self.left_point(self._grid_width_max - points)

    def points_from_bottom(self, points: float) -> int:
        """Returns the pixel position from the specified number of points from
        the bottom border.

        Parameters:
            points (float): The number of grid points from the bottom border

        Returns:
            int: The top pixel value represented by the grid point."""
        self._error_check_top(self._grid_height_max - points)
        return self.top_point(self._grid_height_max - points)

    def draw_grid_to_surface(self, surface: Surface,
                             color: tuple = (0, 0, 0)) -> None:
        """Draws the grid to the pygame surface provided.

        Parameters:
            surface (pygame.Surface): The surface you want to show the grid
                                      on.
            color (tuple): (Optional) Set the color of the grid
                           (default = black (0, 0, 0))."""
        # Draw lines from left to right
        l_point, t_point = 0, 0
        while l_point <= self._grid_width_max:
            draw.line(surface, color,
                      (self.left_point(l_point), self.points_from_top(0)),
                      (self.left_point(l_point), self.points_from_bottom(0)))
            l_point += 1
        # Draw lines from top to bottom
        while t_point <= self._grid_height_max:
            draw.line(surface, color,
                      (self.points_from_left(0), self.top_point(t_point)),
                      (self.points_from_right(0), self.top_point(t_point)))
            t_point += 1



import unittest
import pygame
from pygame_gridcalculator.gridcalculator import GridCalculator
from pygame_gridcalculator.shapefactory import ShapeFactory


class TestShapeFactory(unittest.TestCase):
    def setUp(self) -> None:
        self.test_surface = pygame.Surface((100, 100))
        self.test_grid = GridCalculator(100, 100, 4, 4)
        self.test_shape_factory = ShapeFactory(self.test_grid)

    def test_init(self) -> None:
        """Test initialising ShapeFactory with a grid."""
        result = ShapeFactory(GridCalculator(10, 10, 5, 5))
        self.assertIsInstance(result, ShapeFactory)

    def test_Rect(self) -> None:
        """Test getting a Rect from the test ShapeFactory."""
        result = self.test_shape_factory.Rect(1, 1, 20, 30)
        self.assertIsInstance(result, pygame.Rect)
        self.assertEqual(result.left, 25)
        self.assertEqual(result.top, 25)
        self.assertEqual(result.width, 20)
        self.assertEqual(result.height, 30)

    def test_draw_line(self) -> None:
        """Test getting a pygame.draw.line from the test ShapeFactory"""
        result = self.test_shape_factory.draw_line(self.test_surface,
                                                   (255, 0, 0), (1, 1), (2, 2),
                                                   1)
        self.assertIsInstance(result, pygame.Rect)
        self.assertEqual(result.left, 25)
        self.assertEqual(result.top, 25)
        self.assertEqual(result.width, 26)
        self.assertEqual(result.height, 26)

    def test_draw_lines(self) -> None:
        """Test getting a pygame.draw.lines from the test ShapeFactory"""
        result = self.test_shape_factory.draw_lines(self.test_surface,
                                                    (255, 0, 0), False,
                                                    [(1, 1), (2, 2), (3, 1)],
                                                    1)
        self.assertIsInstance(result, pygame.Rect)
        self.assertEqual(result.left, 25)
        self.assertEqual(result.top, 25)
        self.assertEqual(result.width, 51)
        self.assertEqual(result.height, 26)

    def test_draw_aaline(self) -> None:
        """Test getting a pygame.draw.aaline from the test ShapeFactory"""
        result = self.test_shape_factory.draw_aaline(self.test_surface,
                                                     (255, 0, 0), (2, 2),
                                                     (3, 3), 1)
        self.assertIsInstance(result, pygame.Rect)
        self.assertEqual(result.left, 50)
        self.assertEqual(result.top, 50)
        self.assertEqual(result.width, 26)
        self.assertEqual(result.height, 26)

    def test_draw_aalines(self) -> None:
        """Test getting a pygame.draw.aalines from the test ShapeFactory"""
        result = self.test_shape_factory.draw_aalines(self.test_surface,
                                                      (255, 0, 0), False,
                                                      [(1, 3), (2, 2), (3, 1)],
                                                      1)
        self.assertIsInstance(result, pygame.Rect)
        self.assertEqual(result.left, 25)
        self.assertEqual(result.top, 25)
        self.assertEqual(result.width, 51)
        self.assertEqual(result.height, 51)

    def test_draw_polygon(self) -> None:
        """Test getting a pygame.draw.polygon from the test ShapeFactory"""
        result = self.test_shape_factory.draw_polygon(self.test_surface,
                                                      (255, 0, 0),
                                                      [(2, 2), (3, 3), (4, 3),
                                                       (4, 2)], 1)
        self.assertIsInstance(result, pygame.Rect)
        self.assertEqual(result.left, 50)
        self.assertEqual(result.top, 50)
        self.assertEqual(result.width, 50)
        self.assertEqual(result.height, 26)

    def test_draw_circle(self) -> None:
        """Test getting a pygame.draw.polygon from the test ShapeFactory"""
        result = self.test_shape_factory.draw_circle(self.test_surface,
                                                     (255, 0, 0), (2, 2), 5)
        self.assertIsInstance(result, pygame.Rect)
        self.assertEqual(result.left, 45)
        self.assertEqual(result.top, 45)
        self.assertEqual(result.width, 10)
        self.assertEqual(result.height, 10)



import pygame
from .gridcalculator import GridCalculator


class ShapeFactory:
    def __init__(self, grid_calculator: GridCalculator):
        self.grid = grid_calculator

    def _convert_grid_tuple(self, grid_tuple: tuple) -> tuple:
        """Takes the grid tuple and returns a pixel tuple."""
        left_grid, top_grid = grid_tuple
        return self.grid.left_point(left_grid), self.grid.top_point(top_grid)

    def _convert_grid_list(self, grid_list: list) -> list:
        """Takes a list of grid tuples and returns a list of pixel tuples."""
        points = []
        for grid_point in grid_list:
            points.append(self._convert_grid_tuple(grid_point))
        return points

    # pygame.Rect shapes

    def Rect(self, grid_left: int, grid_top: int, width: float,
             height: float) -> pygame.Rect:
        """Returns a Rect based on grid left and top points.

        Parameters:
            grid_left (int): The left grid point to use.
            grid_top (int): The top grid point to use.
            width (float): The width in pixels.
            height (float): The height in pixels.

        Returns:
            pygame.Rect: A pygame Rect object."""
        return pygame.Rect(self.grid.left_point(grid_left),
                           self.grid.top_point(grid_top),
                           width, height)

    # pygame.draw shapes

    def draw_line(self, surface: pygame.Surface, color: tuple,
                  grid_start_pos: tuple, grid_end_pos: tuple,
                  width: int = 1) -> pygame.Rect:
        """Calls pygame.draw.line based on grid left and top points.

        Parameters:
            surface (pygame.Surface): The surface to draw the line on to.
            color (tuple): The color value for the line.
            grid_start_pos (tuple): The start point in grid positions
                                    (grid_left, grid_top).
            grid_end_pos (tuple): The start point in grid positions
                                  (grid_left, grid_top).
            width (float): (Optional) The width of the line in pixels
                           (default = 1).

        Returns:
            pygame.Rect: A pygame.Rect object."""
        return pygame.draw.line(surface, color,
                                self._convert_grid_tuple(grid_start_pos),
                                self._convert_grid_tuple(grid_end_pos), width)

    def draw_lines(self, surface: pygame.Surface, color: tuple, closed: bool,
                   grid_points: list, width: int = 1) -> pygame.Rect:
        """Calls pygame.draw.lines based on grid left and top points.

        Parameters:
            surface (pygame.Surface): The surface to draw the line on to.
            color (tuple): The color value for the line.
            closed (bool): If True, a line is drawn between the first and last
                           points.
            grid_points (list): The list of points in grid positions
                                (grid_left, grid_top).
            width (int): (Optional) The width of the line (default = 1).

        Returns:
            pygame.Rect: A pygame.Rect object."""
        return pygame.draw.lines(surface, color, closed,
                                 self._convert_grid_list(grid_points), width)

    def draw_aaline(self, surface: pygame.Surface, color: tuple,
                    grid_start_pos: tuple, grid_end_pos: tuple,
                    blend: int = 1) -> pygame.Rect:
        """Calls pygame.draw.aaline based on grid left and top points.

        Parameters:
            surface (pygame.Surface): The surface to draw the line on to.
            color (tuple): The color value for the line.
            grid_start_pos (tuple): The start point in grid positions
                                    (grid_left, grid_top).
            grid_end_pos (tuple): The start point in grid positions
                                  (grid_left, grid_top).
            blend (int): (Optional) The blend of the line (default = 1).

        Returns:
            pygame.Rect: A pygame.Rect object."""
        return pygame.draw.aaline(surface, color,
                                  self._convert_grid_tuple(grid_start_pos),
                                  self._convert_grid_tuple(grid_end_pos),
                                  blend)

    def draw_aalines(self, surface: pygame.Surface, color: tuple, closed: bool,
                     grid_points: list, blend: int = 1):
        """Calls pygame.draw.aalines based on grid left and top points.

        Parameters:
            surface (pygame.Surface): The surface to draw the line on to.
            color (tuple): The color value for the line.
            closed (bool): If True, a line is drawn between the first and last
                           points.
            grid_points (list): The list of points in grid positions
                                (grid_left, grid_top).
            blend (int): (Optional) The blend of the line (default = 1).

        Returns:
            pygame.Rect: A pygame.Rect object."""
        return pygame.draw.aalines(surface, color, closed,
                                   self._convert_grid_list(grid_points), blend)

    def draw_polygon(self, surface: pygame.Surface, color: tuple,
                     grid_points: list, width: int = 0) -> pygame.Rect:
        """Calls pygame.draw.polygon based on grid left and top points.

        Parameters:
            surface (pygame.Surface): The surface to draw the polygon on to.
            color (tuple): The color value for the polygon.
            grid_points (list): A list of grid points in the format
                                (grid_left, grid_top).
            width (int): (Optional) The width of the line (default = 0).

        Returns:
            pygame.Rect: A pygame.Rect object."""
        return pygame.draw.polygon(surface, color,
                                   self._convert_grid_list(grid_points), width)

    def draw_circle(self, surface: pygame.Surface, color: tuple,
                    grid_center: tuple, radius: float, width: int = 0,
                    draw_top_right: bool = True, draw_top_left: bool = True,
                    draw_bottom_left: bool = True,
                    draw_bottom_right: bool = True) -> pygame.Rect:
        """Calls pygame.draw.polygon based on grid left and top points.

        Parameters:
            surface (pygame.Surface): The surface to draw the polygon on to.
            color (tuple): The color value for the polygon.
            grid_center (tuple): The center point in grid positions
                                 (grid_left, grid_top).
            radius (float): The radius of the circle.
            width (int): (Optional) The width of the line (default = 0).
            draw_top_right (bool): (Optional) Specifies if the top right of the
                                   circle should be drawn (default = True).
            draw_top_left (bool): (Optional) Specifies if the top left of the
                                  circle should be drawn (default = True).
            draw_bottom_left (bool): (Optional) Specifies if the top left of
                                     the circle should be drawn
                                     (default = True).
            draw_bottom_right (bool): (Optional) Specifies if the top right of
                                      the circle should be drawn
                                      (default = True).

        Returns:
            pygame.Rect: A pygame.Rect object."""
        return pygame.draw.circle(surface, color,
                                  self._convert_grid_tuple(grid_center),
                                  radius, width, draw_top_right, draw_top_left,
                                  draw_bottom_left, draw_bottom_right)
